aft: Unify handling of hash and path duplicates.
[paraslash.git] / stdout.c
index 4e3b6df9a2582e168c9f144aec44088eb9bc99cc..29db2b7e3e28f050d72577a6d8b533c18ed54be8 100644 (file)
--- a/stdout.c
+++ b/stdout.c
 /*
- * Copyright (C) 2006-2009 Andre Noll <maan@systemlinux.org>
+ * Copyright (C) 2006 Andre Noll <maan@tuebingen.mpg.de>
  *
  * Licensed under the GPL v2. For licencing details see COPYING.
  */
 
 /** \file stdout.c Functions that deal with writing to stdout. */
 
-#include <dirent.h> /* readdir() */
-#include <assert.h>
-#include <stdbool.h>
-
 #include "para.h"
 #include "list.h"
 #include "sched.h"
 #include "fd.h"
 #include "error.h"
 #include "stdout.h"
+#include "buffer_tree.h"
 
-/**
- * The pre_select function of the stdout task.
- *
- * \param s The scheduler this task was registered to.
- * \param t The task structure of the stdout task.
- *
- * This function is always successful. If there is data available in the input
- * buffer, it adds \p STDOUT_FILENO to the write fd set of \a s.
- */
-static void stdout_pre_select(struct sched *s, struct task *t)
+/* Add STDOUT_FILENO to the write fd set if there is input data available. */
+static void stdout_pre_select(struct sched *s, void *context)
 {
-       struct stdout_task *sot = container_of(t, struct stdout_task, task);
+       struct stdout_task *sot = context;
+       int ret;
 
-       t->error = 0;
-       sot->check_fd = 0;
-       if (!*sot->loaded) {
-               if (*sot->input_error < 0) {
-                       t->error = *sot->input_error;
-                       s->timeout.tv_sec = 0;
-                       s->timeout.tv_usec = 1;
-               }
-               return;
-       }
-       sot->check_fd = 1;
-       para_fd_set(STDOUT_FILENO, &s->wfds, &s->max_fileno);
+       ret = btr_node_status(sot->btrn, 0, BTR_NT_LEAF);
+       if (ret > 0)
+               para_fd_set(STDOUT_FILENO, &s->wfds, &s->max_fileno);
+       else if (ret < 0)
+               sched_min_delay(s);
 }
 
-/**
- * The post select function of the stdout task.
- *
- * \param s The scheduler this task was registered to.
- * \param t The task structure of the stdout task.
- *
- * This function checks if \p STDOUT_FILENO was included by in the write fd set
- * of \a s during the previous pre_select call.  If yes, and \p STDOUT_FILENO
- * appeears to be writable, the data loaded in the input buffer is written to
- * stdout.
+/*
+ * This function writes input data from the buffer tree to stdout if
+ * STDOUT_FILENO is writable.
  */
-static void stdout_post_select(struct sched *s, struct task *t)
+static int stdout_post_select(struct sched *s, void *context)
 {
-       struct stdout_task *sot = container_of(t, struct stdout_task, task);
-       ssize_t ret;
+       struct stdout_task *sot = context;
+       struct btr_node *btrn = sot->btrn;
+       int ret;
+       char *buf;
+       size_t sz;
 
-       t->error = 0;
-       if (!sot->check_fd) {
-               if (!*sot->loaded && *sot->input_error < 0)
-                       t->error = *sot->input_error;
-               return;
-       }
+       ret = btr_node_status(btrn, 0, BTR_NT_LEAF);
+       if (ret < 0)
+               goto out;
+       if (ret == 0)
+               return 0;
        if (!FD_ISSET(STDOUT_FILENO, &s->wfds))
-               return;
-       ret = write(STDOUT_FILENO, *sot->bufp, *sot->loaded);
+               return 0;
+
+       if (sot->must_set_nonblock_flag) {
+               ret = mark_fd_nonblocking(STDOUT_FILENO);
+               if (ret < 0)
+                       goto out;
+               sot->must_set_nonblock_flag = false;
+       }
+       for (;;) {
+               sz = btr_next_buffer(btrn, &buf);
+               if (sz == 0)
+                       break;
+               ret = xwrite(STDOUT_FILENO, buf, sz);
+               if (ret <= 0)
+                       break;
+               btr_consume(btrn, ret);
+       }
+out:
        if (ret < 0) {
-               t->error = -ERRNO_TO_PARA_ERROR(errno);
-               return;
+               btr_remove_node(&sot->btrn);
+               /* Revert to blocking mode if necessary. */
+               fcntl(STDOUT_FILENO, F_SETFL, sot->fd_flags);
        }
-       *sot->loaded -= ret;
-       if (*sot->loaded)
-               memmove(*sot->bufp, *sot->bufp + ret, *sot->loaded);
+       return ret;
 }
 
 /**
* Initialize a stdout task structure with default values.
  *
* \param sot The stdout task structure.
  *
- * This fills in the pre/post select function poinzters of the task structure
- * given by \a sot.
+ * This sets up \a sot and registers a task with \a sot as context pointer.
  */
-void stdout_set_defaults(struct stdout_task *sot)
+void stdout_task_register(struct stdout_task *sot, struct sched *s)
 {
        int ret;
+       struct task_info ti = {
+               .pre_select = stdout_pre_select,
+               .post_select = stdout_post_select,
+               .context = sot,
+               .name = "stdout",
+       };
 
-       sot->use_buffer_tree = false;
-       sot->task.pre_select = stdout_pre_select;
-       sot->task.post_select = stdout_post_select;
-       sprintf(sot->task.status, "stdout writer");
-       ret = mark_fd_nonblocking(STDOUT_FILENO);
-       if (ret >= 0)
-               return;
-       PARA_EMERG_LOG("%s\n", para_strerror(-ret));
-       exit(EXIT_FAILURE);
+       /* See stdin.c for details. */
+       ret = fcntl(STDOUT_FILENO, F_GETFL);
+       if (ret < 0) {
+               PARA_EMERG_LOG("F_GETFL: %s\n", strerror(errno));
+               exit(EXIT_FAILURE);
+       }
+       sot->fd_flags = ret;
+       sot->must_set_nonblock_flag = (sot->fd_flags & O_NONBLOCK) == 0;
+       sot->task = task_register(&ti, s);
 }
@@
 
 /**
- * Initialize a stdout task structure with default values.
+ * Register a stdout task structure.
  *
@@
  *
- * \param sot The stdout task structure.
+ * \param sot The stdout task structure to register.
+ * \param s The task will be added to this scheduler's task list.
  *